// Neon++ Library, (C) 2006 M. Derezynski
// A C++ wrapper library for neon <http://www.webdav.org/neon>, (C) Joe Orton

#include <iostream>
#include <istream>
#include <ostream>
#include <string>

#include <ne_session.h>
#include <ne_request.h>

#include "session.hh" 
#include "request.hh" 

namespace Neon
{
  Request::~Request ()
  {
    if (m_request) ne_request_destroy (m_request);
    if (m_session) delete m_session;
  }

  Request::Request (std::string const& hostname,
                    std::string const& path,
                    Neon::HttpPort port,
                    ResponseRead read,
                    Method method,
                    std::string const& scheme)

        throw (Neon::Session::SessionCreateError, InvalidRequestError) : m_request (0), m_session (0), m_dispatched (false)
  {
    m_session = new Neon::Session (hostname, scheme, port);
    m_request = ne_request_create (m_session->session, (method == Neon::Request::METHOD_GET) ? "GET" : "POST", path.c_str());

    if (!m_request)
      throw InvalidRequestError (ne_get_error (m_session->session));

    if (read == Neon::Request::RESPONSE_READ)
    {
      ne_add_response_body_reader (m_request, ne_accept_2xx, read_block, &m_vec);
    }
  }

  void 
  Request::set_request_data (const char * data, size_t data_size)
  {
    ne_set_request_body_buffer (m_request, data, data_size); 
  }

  const ne_status*
  Request::get_status () const
  {
    return ne_get_status (m_request);
  }

  int
  Request::read_block (void * udata, const char * data, size_t len)
  {
    std::vector<unsigned char> *vec = (std::vector<unsigned char>*)(udata);
    for (unsigned int n = 0; n < len; ++n)
      vec->push_back ((unsigned char)data[n]);
    return 0;
  }

  void
  Request::clear ()
  {
    m_dispatched = false;
    m_vec.clear(); 
  }

#ifdef NEONPP_EXCEPTIONS_ENABLED
  void
#else
  int
#endif
  Request::dispatch ()
  {
    if (m_dispatched) return NE_OK;

    int result = ne_request_dispatch (m_request);
    m_dispatched = true;

#ifdef NEONPP_EXCEPTIONS_ENABLED
    switch (result)
    {
      case NE_OK:
        return;

      case NE_AUTH:
        throw NeAuthError (ne_get_error (ne_get_session (m_request)));

      case NE_PROXYAUTH:
        throw NeProxyAuthError (ne_get_error (ne_get_session (m_request)));

      case NE_CONNECT:  
        throw NeConnectError (ne_get_error (ne_get_session (m_request)));

      case NE_ERROR:
        throw NeGenericError (ne_get_error (ne_get_session (m_request)));
    }
#else
    return result;
#endif

  }

  void
  Request::add_request_header (std::string const& name, std::string const& value)
  {
    ne_add_request_header (m_request, name.c_str(), value.c_str());
  }

  void
  Request::add_request_header (Neon::Request::HeaderType type)
  {
    switch (type)
    {
      case HEADER_WWWFORM:
        ne_add_request_header (m_request, "Content-Type", "application/x-www-form-urlencoded");
        break;
    }
  }

  std::ostream&
  operator<<(std::ostream& os, Request & r)
  {
    r.dispatch ();
    os.write (reinterpret_cast<const char*>(&(r.m_vec[0])), r.m_vec.size());
    return os;
  }

  std::ostream&
  Request::operator<<(std::ostream& os)
  {
    dispatch ();
    os.write (reinterpret_cast<const char*>(&(m_vec[0])), m_vec.size());
    return os;
  }

  void
  Request::operator>> (std::string& out)
  {
    dispatch ();
    out.append (reinterpret_cast<const char*>(&m_vec[0])); 
  }

  const unsigned char *
  Request::get_data ()
  {
    return &m_vec[0];
  }

  size_t
  Request::get_data_size ()
  {
    return m_vec.size();
  }

  std::string
  Request::get_error () const
  {
    return std::string (ne_get_error (ne_get_session (m_request)));
  }

}
